@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014 Google Inc.
+ * Copyright 2014 Google LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,20 +15,24 @@
  */
 package com.google.auto.value;
 
-import static org.junit.Assert.assertEquals;
+import static com.google.common.truth.Truth.assertThat;
+import static com.google.common.truth.TruthJUnit.assume;
 
 import com.google.auto.value.annotations.Empty;
 import com.google.auto.value.annotations.GwtArrays;
 import com.google.auto.value.annotations.StringValues;
+import com.google.common.base.StandardSystemProperty;
 import com.google.common.collect.ImmutableCollection;
 import com.google.common.collect.ImmutableList;
 import com.google.common.collect.ImmutableSet;
 import com.google.common.collect.ImmutableSortedSet;
-import com.google.common.primitives.Ints;
 import com.google.common.testing.EqualsTester;
+import com.google.common.testing.SerializableTester;
+import java.io.ObjectStreamClass;
 import java.lang.annotation.Annotation;
 import java.lang.annotation.Retention;
 import java.lang.annotation.RetentionPolicy;
+import java.lang.reflect.Method;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
@@ -38,6 +42,7 @@ import java.util.List;
 import java.util.Set;
 import java.util.SortedSet;
 import java.util.TreeSet;
+import org.junit.AssumptionViolatedException;
 import org.junit.Test;
 import org.junit.runner.RunWith;
 import org.junit.runners.JUnit4;
@@ -65,14 +70,34 @@ public class AutoAnnotationTest {
         .testEquals();
   }
 
+  @Test
+  public void testEqualsParameterAnnotation() throws ReflectiveOperationException {
+    assume()
+        .that(Double.parseDouble(StandardSystemProperty.JAVA_SPECIFICATION_VERSION.value()))
+        .isAtLeast(8.0);
+    Class<? extends Annotation> jspecifyNullable;
+    try {
+      // We write this using .concat in order to hide it from rewriting rules.
+      jspecifyNullable =
+          Class.forName("org".concat(".jspecify.nullness.Nullable")).asSubclass(Annotation.class);
+    } catch (ClassNotFoundException e) {
+      throw new AssumptionViolatedException("No JSpecify @Nullable available", e);
+    }
+    @SuppressWarnings("GetClassOnAnnotation") // yes, I really want the implementation class
+    Class<? extends StringValues> autoAnnotationImpl = newStringValues(new String[0]).getClass();
+    Method equals = autoAnnotationImpl.getDeclaredMethod("equals", Object.class);
+    assertThat(equals.getAnnotatedParameterTypes()[0].isAnnotationPresent(jspecifyNullable))
+        .isTrue();
+  }
+
   @Test
   public void testArraysAreCloned() {
     String[] array = {"Jekyll"};
     StringValues stringValues = newStringValues(array);
     array[0] = "Hyde";
-    assertEquals("Jekyll", stringValues.value()[0]);
+    assertThat(stringValues.value()).asList().containsExactly("Jekyll");
     stringValues.value()[0] = "Hyde";
-    assertEquals("Jekyll", stringValues.value()[0]);
+    assertThat(stringValues.value()[0]).isEqualTo("Jekyll");
   }
 
   @Test
@@ -80,12 +105,12 @@ public class AutoAnnotationTest {
     String[] strings = {"Jekyll"};
     int[] ints = {2, 3, 5};
     GwtArrays arrays = newGwtArrays(strings, ints);
-    assertEquals(ImmutableList.of("Jekyll"), ImmutableList.copyOf(arrays.strings()));
-    assertEquals(ImmutableList.of(2, 3, 5), Ints.asList(arrays.ints()));
+    assertThat(arrays.strings()).asList().containsExactly("Jekyll");
+    assertThat(arrays.ints()).asList().containsExactly(2, 3, 5).inOrder();
     strings[0] = "Hyde";
     ints[0] = -1;
-    assertEquals(ImmutableList.of("Jekyll"), ImmutableList.copyOf(arrays.strings()));
-    assertEquals(ImmutableList.of(2, 3, 5), Ints.asList(arrays.ints()));
+    assertThat(arrays.strings()).asList().containsExactly("Jekyll");
+    assertThat(arrays.ints()).asList().containsExactly(2, 3, 5).inOrder();
   }
 
   @AutoAnnotation
@@ -145,6 +170,8 @@ public class AutoAnnotationTest {
 
     StringValues anAnnotation();
 
+    Class<? extends CharSequence> aClass();
+
     byte[] bytes();
 
     short[] shorts();
@@ -166,6 +193,8 @@ public class AutoAnnotationTest {
     RetentionPolicy[] enums();
 
     StringValues[] annotations();
+
+    Class<? extends CharSequence>[] classes();
   }
 
   @AutoAnnotation
@@ -181,6 +210,7 @@ public class AutoAnnotationTest {
       String aString,
       RetentionPolicy anEnum,
       StringValues anAnnotation,
+      Class<? extends CharSequence> aClass,
       byte[] bytes,
       short[] shorts,
       int[] ints,
@@ -191,7 +221,8 @@ public class AutoAnnotationTest {
       boolean[] booleans,
       String[] strings,
       RetentionPolicy[] enums,
-      StringValues[] annotations) {
+      StringValues[] annotations,
+      Class<? extends CharSequence>... classes) {
     return new AutoAnnotation_AutoAnnotationTest_newEverything(
         aByte,
         aShort,
@@ -204,6 +235,7 @@ public class AutoAnnotationTest {
         aString,
         anEnum,
         anAnnotation,
+        aClass,
         bytes,
         shorts,
         ints,
@@ -214,7 +246,8 @@ public class AutoAnnotationTest {
         booleans,
         strings,
         enums,
-        annotations);
+        annotations,
+        classes);
   }
 
   @AutoAnnotation
@@ -230,6 +263,7 @@ public class AutoAnnotationTest {
       String aString,
       RetentionPolicy anEnum,
       StringValues anAnnotation,
+      Class<? extends CharSequence> aClass,
       Collection<Byte> bytes,
       List<Short> shorts,
       ArrayList<Integer> ints,
@@ -240,7 +274,8 @@ public class AutoAnnotationTest {
       ImmutableCollection<Boolean> booleans,
       ImmutableList<String> strings,
       ImmutableSet<RetentionPolicy> enums,
-      Set<StringValues> annotations) {
+      Set<StringValues> annotations,
+      List<Class<? extends CharSequence>> classes) {
     return new AutoAnnotation_AutoAnnotationTest_newEverythingCollections(
         aByte,
         aShort,
@@ -253,6 +288,7 @@ public class AutoAnnotationTest {
         aString,
         anEnum,
         anAnnotation,
+        aClass,
         bytes,
         shorts,
         ints,
@@ -263,7 +299,8 @@ public class AutoAnnotationTest {
         booleans,
         strings,
         enums,
-        annotations);
+        annotations,
+        classes);
   }
 
   @Everything(
@@ -278,6 +315,7 @@ public class AutoAnnotationTest {
       aString = "maybe\nmaybe not\n",
       anEnum = RetentionPolicy.RUNTIME,
       anAnnotation = @StringValues("whatever"),
+      aClass = String.class,
       bytes = {5, 6},
       shorts = {},
       ints = {7},
@@ -288,7 +326,8 @@ public class AutoAnnotationTest {
       booleans = {false, true, false},
       strings = {"ver", "vers", "vert", "verre", "vair"},
       enums = {RetentionPolicy.CLASS, RetentionPolicy.RUNTIME},
-      annotations = {@StringValues({}), @StringValues({"foo", "bar"})})
+      annotations = {@StringValues({}), @StringValues({"foo", "bar"})},
+      classes = {String.class, StringBuilder.class})
   private static class AnnotatedWithEverything {}
 
   // Get an instance of @Everything via reflection on the class AnnotatedWithEverything,
@@ -309,6 +348,7 @@ public class AutoAnnotationTest {
           "maybe\nmaybe not\n",
           RetentionPolicy.RUNTIME,
           newStringValues(new String[] {"whatever"}),
+          String.class,
           new byte[] {5, 6},
           new short[] {},
           new int[] {7},
@@ -321,7 +361,9 @@ public class AutoAnnotationTest {
           new RetentionPolicy[] {RetentionPolicy.CLASS, RetentionPolicy.RUNTIME},
           new StringValues[] {
             newStringValues(new String[] {}), newStringValues(new String[] {"foo", "bar"}),
-          });
+          },
+          String.class,
+          StringBuilder.class);
   private static final Everything EVERYTHING_FROM_AUTO_COLLECTIONS =
       newEverythingCollections(
           (byte) 1,
@@ -335,6 +377,7 @@ public class AutoAnnotationTest {
           "maybe\nmaybe not\n",
           RetentionPolicy.RUNTIME,
           newStringValues(new String[] {"whatever"}),
+          String.class,
           Arrays.asList((byte) 5, (byte) 6),
           Collections.<Short>emptyList(),
           new ArrayList<Integer>(Collections.singleton(7)),
@@ -347,7 +390,10 @@ public class AutoAnnotationTest {
           ImmutableList.of("ver", "vers", "vert", "verre", "vair"),
           ImmutableSet.of(RetentionPolicy.CLASS, RetentionPolicy.RUNTIME),
           ImmutableSet.of(
-              newStringValues(new String[] {}), newStringValues(new String[] {"foo", "bar"})));
+              newStringValues(new String[] {}), newStringValues(new String[] {"foo", "bar"})),
+          ImmutableList.of(String.class.asSubclass(CharSequence.class), StringBuilder.class));
+          // .asSubclass because of pre-Java8, where otherwise we get a compilation error because
+          // the inferred type is <Class<? extends CharSequence & Serializable>>.
   private static final Everything EVERYTHING_ELSE_FROM_AUTO =
       newEverything(
           (byte) 0,
@@ -361,6 +407,7 @@ public class AutoAnnotationTest {
           "",
           RetentionPolicy.SOURCE,
           newStringValues(new String[] {""}),
+          String.class,
           new byte[0],
           new short[0],
           new int[0],
@@ -385,6 +432,7 @@ public class AutoAnnotationTest {
           "",
           RetentionPolicy.SOURCE,
           newStringValues(new String[] {""}),
+          String.class,
           ImmutableList.<Byte>of(),
           Collections.<Short>emptyList(),
           new ArrayList<Integer>(),
@@ -395,7 +443,8 @@ public class AutoAnnotationTest {
           ImmutableSet.<Boolean>of(),
           ImmutableList.<String>of(),
           ImmutableSet.<RetentionPolicy>of(),
-          Collections.<StringValues>emptySet());
+          Collections.<StringValues>emptySet(),
+          Collections.<Class<? extends CharSequence>>emptyList());
 
   @Test
   public void testEqualsAndHashCode() {
@@ -406,7 +455,39 @@ public class AutoAnnotationTest {
         .testEquals();
   }
 
+  @Test
+  public void testSerialization() {
+    Annotation[] instances = {EVERYTHING_FROM_AUTO, EVERYTHING_FROM_AUTO_COLLECTIONS};
+    for (Annotation instance : instances) {
+      SerializableTester.reserializeAndAssert(instance);
+    }
+  }
+
+  @Test
+  @SuppressWarnings("GetClassOnAnnotation") // yes, we really do want the implementation classes
+  public void testSerialVersionUid() {
+    Class<? extends Everything> everythingImpl = EVERYTHING_FROM_AUTO.getClass();
+    Class<? extends Everything> everythingFromCollectionsImpl =
+        EVERYTHING_FROM_AUTO_COLLECTIONS.getClass();
+    assertThat(everythingImpl).isNotEqualTo(everythingFromCollectionsImpl);
+    long everythingUid = ObjectStreamClass.lookup(everythingImpl).getSerialVersionUID();
+    long everythingFromCollectionsUid =
+        ObjectStreamClass.lookup(everythingFromCollectionsImpl).getSerialVersionUID();
+    // Two different implementations of the same annotation with the same members being provided
+    // (not defaulted) should have the same serialVersionUID. They won't be serial-compatible, of
+    // course, because their classes are different. So we're really just checking that the
+    // serialVersionUID depends only on the names and types of those members.
+    assertThat(everythingFromCollectionsUid).isEqualTo(everythingUid);
+    Class<? extends StringValues> stringValuesImpl = newStringValues(new String[0]).getClass();
+    long stringValuesUid = ObjectStreamClass.lookup(stringValuesImpl).getSerialVersionUID();
+    // The previous assertion would be vacuously true if every implementation had the same
+    // serialVersionUID, so check that that's not true.
+    assertThat(stringValuesUid).isNotEqualTo(everythingUid);
+  }
+
   public static class IntList extends ArrayList<Integer> {
+    private static final long serialVersionUID = 1L;
+
     IntList(Collection<Integer> c) {
       super(c);
     }
@@ -440,7 +521,7 @@ public class AutoAnnotationTest {
     IntList intList = new IntList(ImmutableList.of(1, 2, 3));
     IntArray actual = newIntArray(intList);
     IntArray expected = AnnotatedWithIntArray.class.getAnnotation(IntArray.class);
-    assertEquals(expected, actual);
+    assertThat(actual).isEqualTo(expected);
   }
 
   @Test
@@ -450,6 +531,7 @@ public class AutoAnnotationTest {
             + "aByte=1, aShort=2, anInt=3, aLong=-4, aFloat=NaN, aDouble=NaN, aChar='#', "
             + "aBoolean=true, aString=\"maybe\\nmaybe not\\n\", anEnum=RUNTIME, "
             + "anAnnotation=@com.google.auto.value.annotations.StringValues([\"whatever\"]), "
+            + "aClass=class java.lang.String, "
             + "bytes=[5, 6], shorts=[], ints=[7], longs=[8, 9], floats=[10.0, 11.0], "
             + "doubles=[-Infinity, -12.0, Infinity], "
             + "chars=['?', '!', '\\n'], "
@@ -459,10 +541,11 @@ public class AutoAnnotationTest {
             + "annotations=["
             + "@com.google.auto.value.annotations.StringValues([]), "
             + "@com.google.auto.value.annotations.StringValues([\"foo\", \"bar\"])"
-            + "]"
+            + "], "
+            + "classes=[class java.lang.String, class java.lang.StringBuilder]"
             + ")";
-    assertEquals(expected, EVERYTHING_FROM_AUTO.toString());
-    assertEquals(expected, EVERYTHING_FROM_AUTO_COLLECTIONS.toString());
+    assertThat(EVERYTHING_FROM_AUTO.toString()).isEqualTo(expected);
+    assertThat(EVERYTHING_FROM_AUTO_COLLECTIONS.toString()).isEqualTo(expected);
   }
 
   @Test
@@ -475,7 +558,7 @@ public class AutoAnnotationTest {
     String expected =
         "@com.google.auto.value.annotations.StringValues("
             + "[\"\", \"\\r\\n\", \"hello, world\", \"Éamonn\", \"\\007\\uffef\"])";
-    assertEquals(expected, instance.toString());
+    assertThat(instance.toString()).isEqualTo(expected);
   }
 
   @Retention(RetentionPolicy.RUNTIME)
@@ -498,7 +581,7 @@ public class AutoAnnotationTest {
             ImmutableList.<Class<? extends Annotation>>of(AnnotationsAnnotation.class));
     AnnotationsAnnotation fromReflect =
         AnnotatedWithAnnotationsAnnotation.class.getAnnotation(AnnotationsAnnotation.class);
-    assertEquals(fromReflect, generated);
+    assertThat(generated).isEqualTo(fromReflect);
   }
 
   @Retention(RetentionPolicy.RUNTIME)
@@ -520,7 +603,7 @@ public class AutoAnnotationTest {
         newClassesAnnotation(Arrays.<Class<?>>asList(AnnotationsAnnotation.class));
     ClassesAnnotation fromReflect =
         AnnotatedWithClassesAnnotation.class.getAnnotation(ClassesAnnotation.class);
-    assertEquals(fromReflect, generated);
+    assertThat(generated).isEqualTo(fromReflect);
   }
 
   @Retention(RetentionPolicy.RUNTIME)
