@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012 Google, Inc.
+ * Copyright 2012 Google LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -16,19 +16,23 @@
 package com.google.auto.value.processor;
 
 import static com.google.auto.common.MoreElements.getLocalAndInheritedMethods;
+import static com.google.auto.common.MoreStreams.toImmutableList;
 import static com.google.auto.value.processor.ClassNames.AUTO_VALUE_NAME;
+import static com.google.common.base.Preconditions.checkState;
 import static com.google.common.collect.Sets.difference;
 import static com.google.common.collect.Sets.intersection;
+import static java.util.Comparator.naturalOrder;
 import static java.util.stream.Collectors.joining;
-import static java.util.stream.Collectors.toList;
 
 import com.google.auto.service.AutoService;
 import com.google.auto.value.extension.AutoValueExtension;
 import com.google.common.annotations.VisibleForTesting;
 import com.google.common.base.Strings;
+import com.google.common.base.Throwables;
 import com.google.common.collect.ImmutableBiMap;
 import com.google.common.collect.ImmutableList;
 import com.google.common.collect.ImmutableListMultimap;
+import com.google.common.collect.ImmutableMap;
 import com.google.common.collect.ImmutableSet;
 import java.lang.annotation.Annotation;
 import java.util.ArrayList;
@@ -36,111 +40,146 @@ import java.util.Collections;
 import java.util.HashSet;
 import java.util.List;
 import java.util.Optional;
-import java.util.OptionalInt;
 import java.util.ServiceConfigurationError;
-import java.util.ServiceLoader;
 import java.util.Set;
 import javax.annotation.processing.ProcessingEnvironment;
 import javax.annotation.processing.Processor;
 import javax.annotation.processing.SupportedAnnotationTypes;
-import javax.annotation.processing.SupportedOptions;
 import javax.lang.model.element.AnnotationMirror;
-import javax.lang.model.element.Element;
-import javax.lang.model.element.ElementKind;
 import javax.lang.model.element.ExecutableElement;
 import javax.lang.model.element.TypeElement;
 import javax.lang.model.type.TypeKind;
 import javax.lang.model.type.TypeMirror;
+import net.ltgt.gradle.incap.IncrementalAnnotationProcessor;
+import net.ltgt.gradle.incap.IncrementalAnnotationProcessorType;
 
 /**
  * Javac annotation processor (compiler plugin) for value types; user code never references this
  * class.
  *
- * @see <a href="https://github.com/google/auto/tree/master/value">AutoValue User's Guide</a>
+ * @see <a href="https://github.com/google/auto/tree/main/value">AutoValue User's Guide</a>
  * @author Éamonn McManus
  */
 @AutoService(Processor.class)
 @SupportedAnnotationTypes(AUTO_VALUE_NAME)
-@SupportedOptions("com.google.auto.value.OmitIdentifiers")
-public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
+@IncrementalAnnotationProcessor(IncrementalAnnotationProcessorType.DYNAMIC)
+public class AutoValueProcessor extends AutoValueishProcessor {
+  static final String OMIT_IDENTIFIERS_OPTION = "com.google.auto.value.OmitIdentifiers";
+
+  // We moved MemoizeExtension to a different package, which had an unexpected effect:
+  // now if an old version of AutoValue is in the class path, ServiceLoader can pick up both the
+  // old and the new versions of MemoizeExtension. So we exclude the old version if we see it.
+  // The new version will be bundled with this processor so we should always find it.
+  private static final String OLD_MEMOIZE_EXTENSION =
+      "com.google.auto.value.extension.memoized.MemoizeExtension";
+
   public AutoValueProcessor() {
     this(AutoValueProcessor.class.getClassLoader());
   }
 
   @VisibleForTesting
   AutoValueProcessor(ClassLoader loaderForExtensions) {
-    super(AUTO_VALUE_NAME);
-    this.extensions = null;
-    this.loaderForExtensions = loaderForExtensions;
+    this(ImmutableList.of(), loaderForExtensions);
   }
 
   @VisibleForTesting
-  public AutoValueProcessor(Iterable<? extends AutoValueExtension> extensions) {
-    super(AUTO_VALUE_NAME);
-    this.extensions = ImmutableList.<AutoValueExtension>copyOf(extensions);
-    this.loaderForExtensions = null;
+  public AutoValueProcessor(Iterable<? extends AutoValueExtension> testExtensions) {
+    this(testExtensions, null);
+  }
+
+  private AutoValueProcessor(
+      Iterable<? extends AutoValueExtension> testExtensions, ClassLoader loaderForExtensions) {
+    super(AUTO_VALUE_NAME, /* appliesToInterfaces= */ false);
+    this.extensions = ImmutableList.copyOf(testExtensions);
+    this.loaderForExtensions = loaderForExtensions;
   }
 
   // Depending on how this AutoValueProcessor was constructed, we might already have a list of
-  // extensions when init() is run, or, if `extensions` is null, we have a ClassLoader that will be
-  // used to get the list using the ServiceLoader API.
+  // extensions when init() is run, or, if `loaderForExtensions` is not null, it is a ClassLoader
+  // that will be used to get the list using the ServiceLoader API.
   private ImmutableList<AutoValueExtension> extensions;
   private final ClassLoader loaderForExtensions;
 
+  @VisibleForTesting
+  static ImmutableList<AutoValueExtension> extensionsFromLoader(ClassLoader loader) {
+    return SimpleServiceLoader.load(AutoValueExtension.class, loader).stream()
+        .filter(ext -> !ext.getClass().getName().equals(OLD_MEMOIZE_EXTENSION))
+        .collect(toImmutableList());
+  }
+
   @Override
   public synchronized void init(ProcessingEnvironment processingEnv) {
     super.init(processingEnv);
 
-    if (extensions == null) {
+    if (loaderForExtensions != null) {
+      checkState(extensions.isEmpty());
       try {
-        extensions =
-            ImmutableList.copyOf(ServiceLoader.load(AutoValueExtension.class, loaderForExtensions));
-        // ServiceLoader.load returns a lazily-evaluated Iterable, so evaluate it eagerly now
-        // to discover any exceptions.
-      } catch (Throwable t) {
-        StringBuilder warning = new StringBuilder();
-        warning.append(
-            "An exception occurred while looking for AutoValue extensions. "
-                + "No extensions will function.");
-        if (t instanceof ServiceConfigurationError) {
-          warning.append(" This may be due to a corrupt jar file in the compiler's classpath.");
-        }
-        warning.append(" Exception: ").append(t);
-        errorReporter().reportWarning(warning.toString(), null);
+        extensions = extensionsFromLoader(loaderForExtensions);
+      } catch (RuntimeException | Error e) {
+        String explain =
+            (e instanceof ServiceConfigurationError)
+                ? " This may be due to a corrupt jar file in the compiler's classpath."
+                : "";
+        errorReporter()
+            .reportWarning(
+                null,
+                "[AutoValueExtensionsException] An exception occurred while looking for AutoValue"
+                    + " extensions. No extensions will function.%s\n%s",
+                explain,
+                Throwables.getStackTraceAsString(e));
         extensions = ImmutableList.of();
       }
     }
   }
 
-  private static String generatedSubclassName(TypeElement type, int depth) {
+  @Override
+  public ImmutableSet<String> getSupportedOptions() {
+    ImmutableSet.Builder<String> builder = ImmutableSet.builder();
+    AutoValueExtension.IncrementalExtensionType incrementalType =
+        extensions.stream()
+            .map(e -> e.incrementalType(processingEnv))
+            .min(naturalOrder())
+            .orElse(AutoValueExtension.IncrementalExtensionType.ISOLATING);
+    builder
+        .add(OMIT_IDENTIFIERS_OPTION)
+        .add(Nullables.NULLABLE_OPTION)
+        .addAll(optionsFor(incrementalType));
+    for (AutoValueExtension extension : extensions) {
+      builder.addAll(extension.getSupportedOptions());
+    }
+    return builder.build();
+  }
+
+  private static ImmutableSet<String> optionsFor(
+      AutoValueExtension.IncrementalExtensionType incrementalType) {
+    switch (incrementalType) {
+      case ISOLATING:
+        return ImmutableSet.of(IncrementalAnnotationProcessorType.ISOLATING.getProcessorOption());
+      case AGGREGATING:
+        return ImmutableSet.of(IncrementalAnnotationProcessorType.AGGREGATING.getProcessorOption());
+      case UNKNOWN:
+        return ImmutableSet.of();
+    }
+    throw new AssertionError(incrementalType);
+  }
+
+  static String generatedSubclassName(TypeElement type, int depth) {
     return generatedClassName(type, Strings.repeat("$", depth) + "AutoValue_");
   }
 
   @Override
   void processType(TypeElement type) {
-    if (!hasAnnotationMirror(type, AUTO_VALUE_NAME)) {
-      // This shouldn't happen unless the compilation environment is buggy,
-      // but it has happened in the past and can crash the compiler.
-      errorReporter()
-          .abortWithError(
-              "annotation processor for @AutoValue was invoked with a type"
-                  + " that does not have that annotation; this is probably a compiler bug",
-              type);
-    }
-    if (type.getKind() != ElementKind.CLASS) {
-      errorReporter().abortWithError("@AutoValue only applies to classes", type);
-    }
     if (ancestorIsAutoValue(type)) {
-      errorReporter().abortWithError("One @AutoValue class may not extend another", type);
+      errorReporter()
+          .abortWithError(type, "[AutoValueExtend] One @AutoValue class may not extend another");
     }
     if (implementsAnnotation(type)) {
       errorReporter()
           .abortWithError(
-              "@AutoValue may not be used to implement an annotation"
-                  + " interface; try using @AutoAnnotation instead",
-              type);
+              type,
+              "[AutoValueImplAnnotation] @AutoValue may not be used to implement an annotation"
+                  + " interface; try using @AutoAnnotation or @AutoBuilder instead");
     }
-    checkModifiersIfNested(type);
 
     // We are going to classify the methods of the @AutoValue class into several categories.
     // This covers the methods in the class itself and the ones it inherits from supertypes.
@@ -167,60 +206,101 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
     BuilderSpec builderSpec = new BuilderSpec(type, processingEnv, errorReporter());
     Optional<BuilderSpec.Builder> builder = builderSpec.getBuilder();
     ImmutableSet<ExecutableElement> toBuilderMethods;
+    ImmutableSet<ExecutableElement> builderAbstractMethods;
     if (builder.isPresent()) {
-      toBuilderMethods = builder.get().toBuilderMethods(typeUtils(), abstractMethods);
+      toBuilderMethods = builder.get().toBuilderMethods(typeUtils(), type, abstractMethods);
+      builderAbstractMethods = builder.get().builderAbstractMethods();
     } else {
       toBuilderMethods = ImmutableSet.of();
+      builderAbstractMethods = ImmutableSet.of();
     }
 
-    ImmutableSet<ExecutableElement> propertyMethods =
-        propertyMethodsIn(immutableSetDifference(abstractMethods, toBuilderMethods));
-    ImmutableBiMap<String, ExecutableElement> properties = propertyNameToMethodMap(propertyMethods);
+    ImmutableMap<ExecutableElement, AnnotatedTypeMirror> propertyMethodsAndTypes =
+        propertyMethodsIn(immutableSetDifference(abstractMethods, toBuilderMethods), type);
+    ImmutableMap<String, ExecutableElement> properties =
+        propertyNameToMethodMap(propertyMethodsAndTypes.keySet());
 
     ExtensionContext context =
-        new ExtensionContext(processingEnv, type, properties, abstractMethods);
+        new ExtensionContext(
+            this,
+            processingEnv,
+            type,
+            properties,
+            propertyMethodsAndTypes,
+            abstractMethods,
+            builderAbstractMethods);
     ImmutableList<AutoValueExtension> applicableExtensions = applicableExtensions(type, context);
     ImmutableSet<ExecutableElement> consumedMethods =
         methodsConsumedByExtensions(
             type, applicableExtensions, context, abstractMethods, properties);
+    ImmutableSet<ExecutableElement> consumedBuilderMethods =
+        builderMethodsConsumedByExtensions(
+            type, applicableExtensions, context, builderAbstractMethods);
 
     if (!consumedMethods.isEmpty()) {
       ImmutableSet<ExecutableElement> allAbstractMethods = abstractMethods;
       abstractMethods = immutableSetDifference(abstractMethods, consumedMethods);
       toBuilderMethods = immutableSetDifference(toBuilderMethods, consumedMethods);
-      propertyMethods =
-          propertyMethodsIn(immutableSetDifference(abstractMethods, toBuilderMethods));
-      properties = propertyNameToMethodMap(propertyMethods);
-      context = new ExtensionContext(processingEnv, type, properties, allAbstractMethods);
+      propertyMethodsAndTypes =
+          propertyMethodsIn(immutableSetDifference(abstractMethods, toBuilderMethods), type);
+      properties = propertyNameToMethodMap(propertyMethodsAndTypes.keySet());
+      context =
+          new ExtensionContext(
+              this,
+              processingEnv,
+              type,
+              properties,
+              propertyMethodsAndTypes,
+              allAbstractMethods,
+              builderAbstractMethods);
     }
 
+    ImmutableSet<ExecutableElement> propertyMethods = propertyMethodsAndTypes.keySet();
     boolean extensionsPresent = !applicableExtensions.isEmpty();
     validateMethods(type, abstractMethods, toBuilderMethods, propertyMethods, extensionsPresent);
 
-    String finalSubclass = generatedSubclassName(type, 0);
+    String finalSubclass = TypeSimplifier.simpleNameOf(generatedSubclassName(type, 0));
     AutoValueTemplateVars vars = new AutoValueTemplateVars();
-    vars.finalSubclass = TypeSimplifier.simpleNameOf(finalSubclass);
-    vars.types = processingEnv.getTypeUtils();
-    vars.identifiers =
-        !processingEnv.getOptions().containsKey("com.google.auto.value.OmitIdentifiers");
-    defineSharedVarsForType(type, methods, vars);
-    defineVarsForType(type, vars, toBuilderMethods, propertyMethods, builder);
+    vars.identifiers = !processingEnv.getOptions().containsKey(OMIT_IDENTIFIERS_OPTION);
+    Nullables nullables = Nullables.fromMethods(processingEnv, methods);
+    defineSharedVarsForType(type, methods, nullables, vars);
+    defineVarsForType(
+        type,
+        vars,
+        toBuilderMethods,
+        propertyMethodsAndTypes,
+        builder,
+        nullables,
+        consumedBuilderMethods);
+    vars.builtType = vars.origClass + vars.actualTypes;
+    vars.build = "new " + finalSubclass + vars.actualTypes;
+
+    // If we've encountered problems then we might end up invoking extensions with inconsistent
+    // state. Anyway we probably don't want to generate code which is likely to provoke further
+    // compile errors to add to the ones we've already seen.
+    errorReporter().abortIfAnyError();
 
     GwtCompatibility gwtCompatibility = new GwtCompatibility(type);
     vars.gwtCompatibleAnnotation = gwtCompatibility.gwtCompatibleAnnotationString();
 
+    builder.ifPresent(context::setBuilderContext);
     int subclassDepth = writeExtensions(type, context, applicableExtensions);
     String subclass = generatedSubclassName(type, subclassDepth);
     vars.subclass = TypeSimplifier.simpleNameOf(subclass);
+    vars.finalSubclass = finalSubclass;
     vars.isFinal = (subclassDepth == 0);
     vars.modifiers = vars.isFinal ? "final " : "abstract ";
+    vars.builderClassModifiers =
+        consumedBuilderMethods.isEmpty()
+            ? vars.isFinal ? "static final " : "static "
+            : "abstract static ";
 
     String text = vars.toText();
     text = TypeEncoder.decode(text, processingEnv, vars.pkg, type.asType());
     text = Reformatter.fixup(text);
     writeSourceFile(subclass, text, type);
     GwtSerialization gwtSerialization = new GwtSerialization(gwtCompatibility, processingEnv, type);
-    gwtSerialization.maybeWriteGwtSerializer(vars);
+    gwtSerialization.maybeWriteGwtSerializer(vars, finalSubclass);
   }
 
   // Invokes each of the given extensions to generate its subclass, and returns the number of
@@ -279,9 +359,10 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
       default:
         errorReporter()
             .reportError(
-                "More than one extension wants to generate the final class: "
-                    + finalExtensions.stream().map(this::extensionName).collect(joining(", ")),
-                type);
+                type,
+                "[AutoValueMultiFinal] More than one extension wants to generate the final class:"
+                    + " %s",
+                finalExtensions.stream().map(this::extensionName).collect(joining(", ")));
         break;
     }
     return ImmutableList.copyOf(applicableExtensions);
@@ -292,7 +373,7 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
       ImmutableList<AutoValueExtension> applicableExtensions,
       ExtensionContext context,
       ImmutableSet<ExecutableElement> abstractMethods,
-      ImmutableBiMap<String, ExecutableElement> properties) {
+      ImmutableMap<String, ExecutableElement> properties) {
     Set<ExecutableElement> consumed = new HashSet<>();
     for (AutoValueExtension extension : applicableExtensions) {
       Set<ExecutableElement> consumedHere = new HashSet<>();
@@ -301,11 +382,11 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
         if (propertyMethod == null) {
           errorReporter()
               .reportError(
-                  "Extension "
-                      + extensionName(extension)
-                      + " wants to consume a property that does not exist: "
-                      + consumedProperty,
-                  type);
+                  type,
+                  "[AutoValueConsumeNonexist] Extension %s wants to consume a property that does"
+                      + " not exist: %s",
+                  extensionName(extension),
+                  consumedProperty);
         } else {
           consumedHere.add(propertyMethod);
         }
@@ -314,12 +395,11 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
         if (!abstractMethods.contains(consumedMethod)) {
           errorReporter()
               .reportError(
-                  "Extension "
-                      + extensionName(extension)
-                      + " wants to consume a method that is not one of the abstract methods in this"
-                      + " class: "
-                      + consumedMethod,
-                  type);
+                  type,
+                  "[AutoValueConsumeNotAbstract] Extension %s wants to consume a method that is"
+                      + " not one of the abstract methods in this class: %s",
+                  extensionName(extension),
+                  consumedMethod);
         } else {
           consumedHere.add(consumedMethod);
         }
@@ -327,10 +407,44 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
       for (ExecutableElement repeat : intersection(consumed, consumedHere)) {
         errorReporter()
             .reportError(
-                "Extension "
-                    + extensionName(extension)
-                    + " wants to consume a method that was already consumed by another extension",
-                repeat);
+                repeat,
+                "[AutoValueMultiConsume] Extension %s wants to consume a method that was already"
+                    + " consumed by another extension",
+                extensionName(extension));
+      }
+      consumed.addAll(consumedHere);
+    }
+    return ImmutableSet.copyOf(consumed);
+  }
+
+  private ImmutableSet<ExecutableElement> builderMethodsConsumedByExtensions(
+      TypeElement type,
+      ImmutableList<AutoValueExtension> applicableExtensions,
+      ExtensionContext context,
+      ImmutableSet<ExecutableElement> builderAbstractMethods) {
+    Set<ExecutableElement> consumed = new HashSet<>();
+    for (AutoValueExtension extension : applicableExtensions) {
+      Set<ExecutableElement> consumedHere = new HashSet<>();
+      for (ExecutableElement consumedMethod : extension.consumeBuilderMethods(context)) {
+        if (!builderAbstractMethods.contains(consumedMethod)) {
+          errorReporter()
+              .reportError(
+                  type,
+                  "[AutoValueBuilderConsumeNotAbstract] Extension %s wants to consume a method that"
+                      + " is not one of the abstract methods in this class: %s",
+                  extensionName(extension),
+                  consumedMethod);
+        } else {
+          consumedHere.add(consumedMethod);
+        }
+      }
+      for (ExecutableElement repeat : intersection(consumed, consumedHere)) {
+        errorReporter()
+            .reportError(
+                repeat,
+                "[AutoValueBuilderConsumeNotAbstract] Extension %s wants to consume a method that"
+                    + " was already consumed by another extension",
+                extensionName(extension));
       }
       consumed.addAll(consumedHere);
     }
@@ -352,11 +466,13 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
         // ElementUtils.override that sometimes fails to recognize that one method overrides
         // another, and therefore leaves us with both an abstract method and the subclass method
         // that overrides it. This shows up in AutoValueTest.LukesBase for example.
-        String message = "Abstract method is neither a property getter nor a Builder converter";
-        if (extensionsPresent) {
-          message += ", and no extension consumed it";
-        }
-        errorReporter().reportWarning(message, method);
+        String extensionMessage = extensionsPresent ? ", and no extension consumed it" : "";
+        errorReporter()
+            .reportWarning(
+                method,
+                "[AutoValueBuilderWhat] Abstract method is neither a property getter nor a Builder"
+                    + " converter%s",
+                extensionMessage);
       }
     }
     errorReporter().abortIfAnyError();
@@ -370,21 +486,34 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
       TypeElement type,
       AutoValueTemplateVars vars,
       ImmutableSet<ExecutableElement> toBuilderMethods,
-      ImmutableSet<ExecutableElement> propertyMethods,
-      Optional<BuilderSpec.Builder> builder) {
-    // We can't use ImmutableList.toImmutableList() for obscure Google-internal reasons.
+      ImmutableMap<ExecutableElement, AnnotatedTypeMirror> propertyMethodsAndTypes,
+      Optional<BuilderSpec.Builder> maybeBuilder,
+      Nullables nullables,
+      ImmutableSet<ExecutableElement> consumedBuilderAbstractMethods) {
+    ImmutableSet<ExecutableElement> propertyMethods = propertyMethodsAndTypes.keySet();
     vars.toBuilderMethods =
-        ImmutableList.copyOf(toBuilderMethods.stream().map(SimpleMethod::new).collect(toList()));
-    ImmutableBiMap<ExecutableElement, String> methodToPropertyName =
-        propertyNameToMethodMap(propertyMethods).inverse();
+        toBuilderMethods.stream().map(SimpleMethod::new).collect(toImmutableList());
+    vars.toBuilderConstructor = !vars.toBuilderMethods.isEmpty();
+    ImmutableListMultimap<ExecutableElement, AnnotationMirror> annotatedPropertyFields =
+        propertyFieldAnnotationMap(type, propertyMethods);
     ImmutableListMultimap<ExecutableElement, AnnotationMirror> annotatedPropertyMethods =
         propertyMethodAnnotationMap(type, propertyMethods);
-    vars.props = propertySet(type, propertyMethods, annotatedPropertyMethods);
-    vars.serialVersionUID = getSerialVersionUID(type);
+    vars.props =
+        propertySet(
+            propertyMethodsAndTypes,
+            annotatedPropertyFields,
+            annotatedPropertyMethods,
+            nullables);
     // Check for @AutoValue.Builder and add appropriate variables if it is present.
-    if (builder.isPresent()) {
-      builder.get().defineVars(vars, methodToPropertyName);
-    }
+    maybeBuilder.ifPresent(
+        builder -> {
+          ImmutableBiMap<ExecutableElement, String> methodToPropertyName =
+              propertyNameToMethodMap(propertyMethods).inverse();
+          builder.defineVarsForAutoValue(
+              vars, methodToPropertyName, nullables, consumedBuilderAbstractMethods);
+          vars.builderName = "Builder";
+          vars.builderAnnotations = copiedClassAnnotations(builder.builderType());
+        });
   }
 
   @Override
@@ -392,47 +521,6 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
     return nullableAnnotationFor(propertyMethod, propertyMethod.getReturnType());
   }
 
-  /**
-   * Returns an appropriate annotation spelling to indicate the nullability of an element. If the
-   * return value is a non-empty Optional, that indicates that the element is nullable, and the
-   * string should be used to annotate it. If the return value is an empty Optional, the element is
-   * not nullable. The return value can be {@code Optional.of("")}, which indicates that the element
-   * is nullable but that the nullability comes from a type annotation. In this case, the annotation
-   * will appear when the type is written, and must not be specified again. If the Optional contains
-   * a present non-empty string then that string will end with a space.
-   *
-   * @param element the element that might be {@code @Nullable}, either a method or a parameter.
-   * @param elementType the relevant type of the element: the return type for a method, or the
-   *     parameter type for a parameter.
-   */
-  static Optional<String> nullableAnnotationFor(Element element, TypeMirror elementType) {
-    List<? extends AnnotationMirror> typeAnnotations = elementType.getAnnotationMirrors();
-    if (nullableAnnotationIndex(typeAnnotations).isPresent()) {
-      return Optional.of("");
-    }
-    List<? extends AnnotationMirror> elementAnnotations = element.getAnnotationMirrors();
-    OptionalInt nullableAnnotationIndex = nullableAnnotationIndex(elementAnnotations);
-    if (nullableAnnotationIndex.isPresent()) {
-      ImmutableList<String> annotations = annotationStrings(elementAnnotations);
-      return Optional.of(annotations.get(nullableAnnotationIndex.getAsInt()) + " ");
-    } else {
-      return Optional.empty();
-    }
-  }
-
-  private static OptionalInt nullableAnnotationIndex(List<? extends AnnotationMirror> annotations) {
-    for (int i = 0; i < annotations.size(); i++) {
-      if (isNullable(annotations.get(i))) {
-        return OptionalInt.of(i);
-      }
-    }
-    return OptionalInt.empty();
-  }
-
-  private static boolean isNullable(AnnotationMirror annotation) {
-    return annotation.getAnnotationType().asElement().getSimpleName().contentEquals("Nullable");
-  }
-
   static ImmutableSet<ExecutableElement> prefixedGettersIn(Iterable<ExecutableElement> methods) {
     ImmutableSet.Builder<ExecutableElement> getters = ImmutableSet.builder();
     for (ExecutableElement method : methods) {
@@ -476,7 +564,7 @@ public class AutoValueProcessor extends AutoValueOrOneOfProcessor {
     if (Collections.disjoint(a, b)) {
       return a;
     } else {
-      return ImmutableSet.copyOf(difference(a, b));
+      return difference(a, b).immutableCopy();
     }
   }
 }
