@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014 Google Inc.
+ * Copyright 2014 Google LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,24 +15,13 @@
  */
 package com.google.auto.value.processor;
 
-import static com.google.common.truth.Truth.assertAbout;
 import static com.google.common.truth.Truth.assertThat;
-import static com.google.common.truth.Truth.assert_;
-import static com.google.testing.compile.JavaSourcesSubjectFactory.javaSources;
-import static org.junit.Assert.assertTrue;
+import static com.google.testing.compile.CompilationSubject.assertThat;
+import static com.google.testing.compile.Compiler.javac;
 
-import com.google.common.collect.ImmutableList;
+import com.google.testing.compile.Compilation;
 import com.google.testing.compile.JavaFileObjects;
-import java.io.File;
-import java.io.IOException;
-import java.io.Writer;
-import java.util.List;
-import javax.tools.Diagnostic;
-import javax.tools.DiagnosticCollector;
-import javax.tools.JavaCompiler;
-import javax.tools.JavaFileManager;
 import javax.tools.JavaFileObject;
-import javax.tools.ToolProvider;
 import org.junit.Test;
 import org.junit.runner.RunWith;
 import org.junit.runners.JUnit4;
@@ -85,10 +74,13 @@ public class AutoAnnotationCompilationTest {
             "",
             "import com.example.annotations.MyAnnotation;",
             "import com.example.enums.MyEnum;",
+            "import java.io.Serializable;",
             GeneratedImport.importGeneratedAnnotationType(),
             "",
             "@Generated(\"" + AutoAnnotationProcessor.class.getName() + "\")",
-            "final class AutoAnnotation_AnnotationFactory_newMyAnnotation implements MyAnnotation {",
+            "final class AutoAnnotation_AnnotationFactory_newMyAnnotation",
+            "     implements MyAnnotation, Serializable {",
+            "  private static final long serialVersionUID = -7473814294717163169L;",
             "  private final MyEnum value;",
             "  private static final int defaultedValue = 23;",
             "",
@@ -123,7 +115,7 @@ public class AutoAnnotationCompilationTest {
             "    }",
             "    if (o instanceof MyAnnotation) {",
             "      MyAnnotation that = (MyAnnotation) o;",
-            "      return (value.equals(that.value()))",
+            "      return value.equals(that.value())",
             "          && (defaultedValue == that.defaultedValue());",
             "    }",
             "    return false;",
@@ -132,17 +124,20 @@ public class AutoAnnotationCompilationTest {
             "  @Override public int hashCode() {",
             "    return ",
             "        " + invariableHash,
-            "        + (" + 127 * "value".hashCode() + " ^ (value.hashCode()))",
+            "        + (" + 127 * "value".hashCode() + " ^ value.hashCode())",
             "    ;",
             "  }",
             "}");
-    assert_()
-        .about(javaSources())
-        .that(ImmutableList.of(annotationFactoryJavaFile, myAnnotationJavaFile, myEnumJavaFile))
-        .processedWith(new AutoAnnotationProcessor())
-        .compilesWithoutError()
-        .and()
-        .generatesSources(expectedOutput);
+    Compilation compilation =
+        javac()
+            .withProcessors(new AutoAnnotationProcessor())
+            .withOptions("-A" + Nullables.NULLABLE_OPTION + "=")
+            .compile(annotationFactoryJavaFile, myAnnotationJavaFile, myEnumJavaFile);
+    assertThat(compilation).succeededWithoutWarnings();
+    assertThat(compilation)
+        .generatedSourceFile(
+            "com.example.factories.AutoAnnotation_AnnotationFactory_newMyAnnotation")
+        .hasSourceEquivalentTo(expectedOutput);
   }
 
   @Test
@@ -165,10 +160,13 @@ public class AutoAnnotationCompilationTest {
     JavaFileObject expectedOutput =
         JavaFileObjects.forSourceLines(
             "AutoAnnotation_AnnotationFactory_newMyAnnotation",
+            "import java.io.Serializable;",
             GeneratedImport.importGeneratedAnnotationType(),
             "",
             "@Generated(\"" + AutoAnnotationProcessor.class.getName() + "\")",
-            "final class AutoAnnotation_AnnotationFactory_newMyAnnotation implements MyAnnotation {",
+            "final class AutoAnnotation_AnnotationFactory_newMyAnnotation",
+            "    implements MyAnnotation, Serializable {",
+            "  private static final long serialVersionUID = 0L;",
             "  AutoAnnotation_AnnotationFactory_newMyAnnotation() {",
             "  }",
             "",
@@ -195,12 +193,15 @@ public class AutoAnnotationCompilationTest {
             "    return 0;",
             "  }",
             "}");
-    assertAbout(javaSources())
-        .that(ImmutableList.of(annotationFactoryJavaFile, myAnnotationJavaFile))
-        .processedWith(new AutoAnnotationProcessor())
-        .compilesWithoutError()
-        .and()
-        .generatesSources(expectedOutput);
+    Compilation compilation =
+        javac()
+            .withProcessors(new AutoAnnotationProcessor())
+            .withOptions("-A" + Nullables.NULLABLE_OPTION + "=")
+            .compile(annotationFactoryJavaFile, myAnnotationJavaFile);
+    assertThat(compilation).succeededWithoutWarnings();
+    assertThat(compilation)
+        .generatedSourceFile("AutoAnnotation_AnnotationFactory_newMyAnnotation")
+        .hasSourceEquivalentTo(expectedOutput);
   }
 
   @Test
@@ -242,11 +243,14 @@ public class AutoAnnotationCompilationTest {
             "package com.example.factories;",
             "",
             "import com.example.annotations.MyAnnotation;",
+            "import java.io.Serializable;",
             "import java.util.Arrays;",
             GeneratedImport.importGeneratedAnnotationType(),
             "",
             "@Generated(\"" + AutoAnnotationProcessor.class.getName() + "\")",
-            "final class AutoAnnotation_AnnotationFactory_newMyAnnotation implements MyAnnotation {",
+            "final class AutoAnnotation_AnnotationFactory_newMyAnnotation implements MyAnnotation,"
+                + " Serializable {",
+            "  private static final long serialVersionUID = -8116050813861599066L;",
             "  private final int[] value;",
             "",
             "  AutoAnnotation_AnnotationFactory_newMyAnnotation(int[] value) {",
@@ -276,28 +280,29 @@ public class AutoAnnotationCompilationTest {
             "    }",
             "    if (o instanceof MyAnnotation) {",
             "      MyAnnotation that = (MyAnnotation) o;",
-            "      return (Arrays.equals(value,",
+            "      return Arrays.equals(value,",
             "          (that instanceof AutoAnnotation_AnnotationFactory_newMyAnnotation)",
             "              ? ((AutoAnnotation_AnnotationFactory_newMyAnnotation) that).value",
-            "              : that.value()));",
+            "              : that.value());",
             "    }",
             "    return false;",
             "  }",
             "",
             "  @Override public int hashCode() {",
             "    return ",
-            "        + (" + 127 * "value".hashCode() + " ^ (Arrays.hashCode(value)));",
+            "        + (" + 127 * "value".hashCode() + " ^ Arrays.hashCode(value));",
             "  }",
             "}");
-    assert_()
-        .about(javaSources())
-        .that(
-            ImmutableList.of(
-                annotationFactoryJavaFile, myAnnotationJavaFile, gwtCompatibleJavaFile))
-        .processedWith(new AutoAnnotationProcessor())
-        .compilesWithoutError()
-        .and()
-        .generatesSources(expectedOutput);
+    Compilation compilation =
+        javac()
+            .withProcessors(new AutoAnnotationProcessor())
+            .withOptions("-A" + Nullables.NULLABLE_OPTION + "=")
+            .compile(annotationFactoryJavaFile, myAnnotationJavaFile, gwtCompatibleJavaFile);
+    assertThat(compilation).succeededWithoutWarnings();
+    assertThat(compilation)
+        .generatedSourceFile(
+            "com.example.factories.AutoAnnotation_AnnotationFactory_newMyAnnotation")
+        .hasSourceEquivalentTo(expectedOutput);
   }
 
   @Test
@@ -347,6 +352,7 @@ public class AutoAnnotationCompilationTest {
             "",
             "import com.example.annotations.MyAnnotation;",
             "import com.example.enums.MyEnum;",
+            "import java.io.Serializable;",
             "import java.util.Arrays;",
             "import java.util.Collection;",
             "import java.util.List;",
@@ -354,7 +360,9 @@ public class AutoAnnotationCompilationTest {
             GeneratedImport.importGeneratedAnnotationType(),
             "",
             "@Generated(\"" + AutoAnnotationProcessor.class.getName() + "\")",
-            "final class AutoAnnotation_AnnotationFactory_newMyAnnotation implements MyAnnotation {",
+            "final class AutoAnnotation_AnnotationFactory_newMyAnnotation implements MyAnnotation,"
+                + " Serializable {",
+            "  private static final long serialVersionUID = -2102364343628921304L;",
             "  private final int[] value;",
             "  private final MyEnum[] enums;",
             "",
@@ -368,7 +376,7 @@ public class AutoAnnotationCompilationTest {
             "    if (enums == null) {",
             "      throw new NullPointerException(\"Null enums\");",
             "    }",
-            "    this.enums = enums.toArray(new MyEnum[0];",
+            "    this.enums = enums.toArray(new MyEnum[0]);",
             "  }",
             "",
             "  @Override public Class<? extends MyAnnotation> annotationType() {",
@@ -399,22 +407,22 @@ public class AutoAnnotationCompilationTest {
             "    }",
             "    if (o instanceof MyAnnotation) {",
             "      MyAnnotation that = (MyAnnotation) o;",
-            "      return (Arrays.equals(value,",
+            "      return Arrays.equals(value,",
             "          (that instanceof AutoAnnotation_AnnotationFactory_newMyAnnotation)",
             "              ? ((AutoAnnotation_AnnotationFactory_newMyAnnotation) that).value",
-            "              : that.value()))",
-            "          && (Arrays.equals(enums,",
+            "              : that.value())",
+            "          && Arrays.equals(enums,",
             "          (that instanceof AutoAnnotation_AnnotationFactory_newMyAnnotation)",
             "              ? ((AutoAnnotation_AnnotationFactory_newMyAnnotation) that).enums",
-            "              : that.enums()))",
+            "              : that.enums());",
             "    }",
             "    return false;",
             "  }",
             "",
             "  @Override public int hashCode() {",
             "    return ",
-            "        + (" + 127 * "value".hashCode() + " ^ (Arrays.hashCode(value)))",
-            "        + (" + 127 * "enums".hashCode() + " ^ (Arrays.hashCode(enums)));",
+            "        + (" + 127 * "value".hashCode() + " ^ Arrays.hashCode(value))",
+            "        + (" + 127 * "enums".hashCode() + " ^ Arrays.hashCode(enums));",
             "  }",
             "",
             "  private static int[] intArrayFromCollection(Collection<Integer> c) {",
@@ -426,28 +434,20 @@ public class AutoAnnotationCompilationTest {
             "    return a;",
             "  }",
             "}");
-    assert_()
-        .about(javaSources())
-        .that(ImmutableList.of(annotationFactoryJavaFile, myEnumJavaFile, myAnnotationJavaFile))
-        .processedWith(new AutoAnnotationProcessor())
-        .compilesWithoutError()
-        .and()
-        .generatesSources(expectedOutput);
+    Compilation compilation =
+        javac()
+            .withProcessors(new AutoAnnotationProcessor())
+            .withOptions("-A" + Nullables.NULLABLE_OPTION + "=")
+            .compile(annotationFactoryJavaFile, myEnumJavaFile, myAnnotationJavaFile);
+    assertThat(compilation).succeededWithoutWarnings();
+    assertThat(compilation)
+        .generatedSourceFile(
+            "com.example.factories.AutoAnnotation_AnnotationFactory_newMyAnnotation")
+        .hasSourceEquivalentTo(expectedOutput);
   }
 
   @Test
-  public void testMissingClass() throws IOException {
-    File tempDir = File.createTempFile("AutoAnnotationCompilationTest", "");
-    assertTrue(tempDir.delete());
-    assertTrue(tempDir.mkdir());
-    try {
-      doTestMissingClass(tempDir);
-    } finally {
-      removeDirectory(tempDir);
-    }
-  }
-
-  private void doTestMissingClass(File tempDir) {
+  public void testMissingClass() {
     // Test that referring to an undefined annotation does not trigger @AutoAnnotation processing.
     // The class Erroneous references an undefined annotation @NotAutoAnnotation. If we didn't have
     // any special treatment of undefined types then we could run into a compiler bug where
@@ -468,34 +468,16 @@ public class AutoAnnotationCompilationTest {
             "  @AutoAnnotation static Empty newEmpty() {}",
             "  @NotAutoAnnotation Empty notNewEmpty() {}",
             "}");
-    JavaCompiler javaCompiler = ToolProvider.getSystemJavaCompiler();
-    DiagnosticCollector<JavaFileObject> diagnosticCollector =
-        new DiagnosticCollector<JavaFileObject>();
-    JavaCompiler.CompilationTask compilationTask =
-        javaCompiler.getTask(
-            (Writer) null,
-            (JavaFileManager) null,
-            diagnosticCollector,
-            ImmutableList.of("-d", tempDir.toString()),
-            (Iterable<String>) null,
-            ImmutableList.of(erroneousJavaFileObject));
-    compilationTask.setProcessors(ImmutableList.of(new AutoAnnotationProcessor()));
-    boolean result = compilationTask.call();
-    assertThat(result).isFalse();
-    List<Diagnostic<? extends JavaFileObject>> diagnostics = diagnosticCollector.getDiagnostics();
-    assertThat(diagnostics).isNotEmpty();
-    assertThat(diagnostics.get(0).getMessage(null)).contains("NotAutoAnnotation");
-    assertThat(diagnostics.get(0).getMessage(null)).doesNotContain("static");
-  }
-
-  private static void removeDirectory(File dir) {
-    for (File file : dir.listFiles()) {
-      if (file.isDirectory()) {
-        removeDirectory(file);
-      } else {
-        assertTrue(file.delete());
-      }
-    }
-    assertTrue(dir.delete());
+    Compilation compilation =
+        javac().withProcessors(new AutoAnnotationProcessor()).compile(erroneousJavaFileObject);
+    assertThat(compilation)
+        .hadErrorContaining("NotAutoAnnotation")
+        .inFile(erroneousJavaFileObject)
+        .onLineContaining("@NotAutoAnnotation");
+    assertThat(
+            compilation.errors().stream()
+                .map(diag -> diag.getMessage(null))
+                .filter(m -> m.contains("static")))
+        .isEmpty();
   }
 }
